﻿; ------------------------------------------------------------------------------------
; Job Queue for Multi - Threading
; Author: Alexander Aigner
; PB 4.3+ x86/x64 Ascii/Unicode
; ------------------------------------------------------------------------------------

; EnableExplicit

; ------------------------------------------------------------------------------------
; Compilercheck
; ------------------------------------------------------------------------------------

CompilerIf #PB_Compiler_Thread = #False
  CompilerError "Enable Threadsafe"  
CompilerEndIf

CompilerIf (#PB_Compiler_Processor <> #PB_Processor_x86) And (#PB_Compiler_Processor <> #PB_Processor_x64)
  CompilerError "Unsupported Processor"
CompilerEndIf

; ------------------------------------------------------------------------------------
; Prototypes, Structures & Enumerations
; ------------------------------------------------------------------------------------

Prototype JobQueue_Job(*JobData)

Structure _JobQueue_JobInfo
  ReturnValue.i ; Rückgabewert
  State.i ; Job Status (Finished/InProgress)
  *Childs.tStack ; Child - Jobs
  *SendFinishSignal ; Finish - Signal senden
EndStructure

Structure _JobQueue_Job
  *Call.JobQueue_Job ; Job - Procedure
  *Info._JobQueue_JobInfo ; Job RückgabeInfo
  Para.i ; Übergabeparameter
  RealTime.i ; Job benötigt Echtzeit - Priorität
  *Parent._JobQueue_JobInfo ; Parent - Job
EndStructure

Structure _JobQueue_Worker
  Id.i ; ThreadID
  JobsTotal.i ; Anzahl fertiger Jobs
  Number.i ; Worker Number
  DoCheck.i ; Check - Signal
  Mutex.i ; Worker Mutex
EndStructure

Structure _JobQueue_WorkerHolder
  Entry._JobQueue_Worker[0]
EndStructure

Structure _JobQueue_Data
  *Thread._JobQueue_WorkerHolder ; Worker Threads
  *Cleaner ; Result - Cleaner
  Mutex.i ; Kontroll Mutex
  DelayMutex.i ; Mutex für die Delay - Zugriffe
  *Stack.tStack[3] ; Stacks für 3 Prioritätsklassen
  *ReturnBuffer.tStack ; Speicher für die Rückgabewerte
  Stop.i ; Deinit - Flag (= ThreadCount: Kill)
  Stoped.i ; Deinit - AnswerFlag
  Pause.i ; Pause - Flag
  Resume.i ; Resume - Signal
  CleanDelay.i ; Clean - Delay
  WorkerCount.i ; Anzahl Worker
  JobCount.i ; Ausstehende Jobs
  JobDelayed.i ; Jobs die noch in der Wartezeit sind
  JobsinProgress.i ; Jobs in Ausführung
  WaitForFinish.i ; Wartet auf Fertigstellung
  AllJobsFinished.i ; Alles Fertig - Signal
EndStructure

Structure _JobQueue_WorkerInit
  *JData._JobQueue_Data
  WorkerNumber.i ;Thread Nummer
EndStructure

Structure _JobQueue_DelayLauncher
  *JobQueue._JobQueue_Data
  *Job._JobQueue_Job
  Priority.i
EndStructure

Enumeration 0 ; Priority - Klassen
  #Job_LowPriority ; Niedrige Priorität, wird zuletzt bearbeitet
  #Job_NormalPriority ; Normale Prioriät, vor Low, hinter High
  #Job_HighPriority ; Hohe Priorität, wird zuerst behandelt
  #Job_HighestPriority ; Höchste Priorität: Wird an erste Stelle von High Priority gepusht
  #Job_RealTimePriority ; Höchste Priorität + Echtzeit!!!! Nur benutzen wenn Sie wissen was sie tun
EndEnumeration

Enumeration 1 ; Messages
  #Worker_Quit ; Thread beenden
  #Worker_Sleep ; Thread Warten lassen
EndEnumeration

Enumeration 1 ; Job Statuszustände
  #Job_inProgress ; in Arbeit
  #Job_Finished ; Fertig
  #Job_Checked ; Job-Rückgabewert wurde abgeholt
EndEnumeration

; ------------------------------------------------------------------------------------
; Internal
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
; Stack Funktionen
; ------------------------------------------------------------------------------------

XIncludeFile "common\Stack.pbi"
XIncludeFile "common\ET_Timer.pbi"

; ------------------------------------------------------------------------------------
; Worker Internal
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
; Declares
; ------------------------------------------------------------------------------------

Declare JQ_Pause(*JobQueue._JobQueue_Data, WaitUntilConfirmation = #False)
Declare JQ_Resume(*JobQueue._JobQueue_Data)
Declare JQ_ClearReturnBuffer(*JobQueue._JobQueue_Data, RemoveUncheckedJobs = #True)
Declare JQ_Add(*JobQueue._JobQueue_Data, *JobProcedure.JobQueue_Job, Parameter.i, Priority = #Job_NormalPriority, Parent = #Null, MinimumLaunchDelay = #Null, _DelayLaunchInfo = #Null)

; ist ein Check-Signal da?
Macro _JQ_CheckForCheckSignal()
  If *Worker\DoCheck
    LockMutex(*Worker\Mutex)
    *Worker\DoCheck = #False
    UnlockMutex(*Worker\Mutex)
    Break 2
  EndIf
EndMacro

; Sendet ein Check-Signal an einen Worker
Macro _JQ_SendCheckSignalToWorker(_JobQueue, _Worker)
  LockMutex(_JobQueue\Thread\Entry[_Worker]\Mutex)
  _JobQueue\Thread\Entry[_Worker]\DoCheck = #True
  UnlockMutex(_JobQueue\Thread\Entry[_Worker]\Mutex)
EndMacro

; Sendet ein Check-Signal an alle Worker
Macro _JQ_SendCheckSignal(_JobQueue)
  Protected cWorker
  For cWorker = 1 To _JobQueue\WorkerCount
    _JQ_SendCheckSignalToWorker(_JobQueue, cWorker)
  Next
EndMacro

; Räumt den JobResult - Stack in regelmäßigen Abständen auf
Procedure JQ_JobResultCleaner(*JobQueue._JobQueue_Data)
    JQ_ClearReturnBuffer(*JobQueue, #False)
EndProcedure

; Führt einen Job verzögert aus.
Procedure JQ_DelayLauncher(*Launcher._JobQueue_DelayLauncher)
  With *Launcher\Job
    JQ_Add(*Launcher\JobQueue, \Call, \Para, *Launcher\Priority, \Parent, #Null, \Info)
  EndWith
 ; FreeMemory(*Launcher)
EndProcedure

; Worker Thread: Arbeitet die Jobs ab
; Der Worker bestimmt wie schnell die gesamte Queue ist, je weniger Overhead desto besser

Procedure JQ_Worker(*Init._JobQueue_WorkerInit)
  Protected oldPriority, *ParentJob._JobQueue_JobInfo
  Protected *Job._JobQueue_Job, Priority
  Protected *JobQueue._JobQueue_Data = *Init\JData
  Protected WorkerNumber = *Init\WorkerNumber
  Protected *Worker._JobQueue_Worker = *JobQueue\Thread\Entry[*Init\WorkerNumber]
  Protected Mutex = *JobQueue\Mutex
  
  FreeMemory(*Init) ; Brauchen wir nicht mehr
  
  With *JobQueue
    Repeat ; Und los gehts
      Delay(1) ; CPU schonen
      
      If \Stop ; JobQueue entfernen
        LockMutex(Mutex)
        \Stoped + 1 ; Ein Worker weniger
        UnlockMutex(Mutex)
        ProcedureReturn #True ; beendet sich selbst
      ElseIf \Pause ; Workers Pausieren
        SignalSemaphore(\Pause)
        WaitSemaphore(\Resume) ; Warten auf Resume Signal
      EndIf
      
      ; Hier könnte bei zu vielen Threads das Reader-Writer Problem zu Verzögerungen
      ; durch zu viele Mutex Anfragen führen, dadurch das \JobsinProgress abgezogen
      ; wird minimiert sich das aber.
      
      If \JobCount-\JobsinProgress ; Ausführen wenn keine Pause ist und Jobs da sind
        LockMutex(Mutex) ; Mutex sperren
        For Priority = #Job_HighPriority To #Job_LowPriority Step -1 ; Höchste zu niedrigste Priorität
          While Stack_Count(\Stack[Priority]) ; Jobs abarbeiten
            _JQ_CheckForCheckSignal() ; Worker soll Flags & seine Messages checken
            
            \JobsinProgress + 1 ; möglichs früh setzen um das Reader-Writer Problem klein zu halten
            
            *Job = Stack_Pop(\Stack[Priority])
            
            If *Job\Parent ; Job hat einen Parent
              *ParentJob = *Job\Parent
              If *ParentJob\State<#Job_Finished ; Parent noch nicht fertig
                If Not *ParentJob\Childs
                  *ParentJob\Childs = Stack_New()
                EndIf
                Stack_Push(*ParentJob\Childs, *Job) ; Job aus der Reihe nehmen und an seinen Vater heften
                \JobsinProgress-1
                Continue ; Weiter mit dem nächsten
              EndIf
            EndIf
            
            If *Job\RealTime ; ECHTZEIT Priorität, ACHTUNG !!!
              oldPriority = ThreadPriority(*Worker\Id, 32)
            EndIf
            
            *Job\Info\State = #Job_inProgress ; Job in Arbeit
            UnlockMutex(Mutex) ; während des Calls muss die Mutex nicht gelockt sein
            
            *Job\Info\ReturnValue = *Job\Call(*Job\Para) ; Job ausführen und Rückgabe speichern
            
            LockMutex(Mutex) ; Mutex wieder sperren
            *Job\Info\State = #Job_Finished ; Job fertig
            
            If *Job\Info\Childs ; Wir haben Kinder bekommen!
              While Stack_Count(*Job\Info\Childs) ; Alle Kinder wieder rein...
                Stack_Push(\Stack[Priority], Stack_Pop(*Job\Info\Childs)) ; Kind zurück in die Umlaufbahn
              Wend
              Stack_Free(*Job\Info\Childs)
            EndIf
            
            \JobCount-1; Einer mehr erledigt
            \JobsinProgress-1
            *Worker\JobsTotal + 1
            
            If *Job\Info\SendFinishSignal ; JobFinish Signal senden
              SignalSemaphore(*Job\Info\SendFinishSignal)
            EndIf
            
            If *Job\RealTime
              ThreadPriority(*Worker\Id, oldPriority)
            EndIf
            
            FreeMemory(*Job) ; Job Memory freigeben (Info bleiben erhalten, da eigener Speicher)
          Wend
        Next
        If Not \JobCount ; AllesFertig - Signal senden
          If \WaitForFinish
            \WaitForFinish = #False
            SignalSemaphore(\AllJobsFinished)
          EndIf
        EndIf
        UnlockMutex(Mutex) ; Mutex freigeben
      EndIf
    ForEver
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
; Public Functions
; ------------------------------------------------------------------------------------

; Initialisiert eine neue Job Queue, Worker = Anzahl Threads
Procedure JQ_Init(Worker = 1)
  Protected i, *JobQueue._JobQueue_Data = AllocateMemory(SizeOf(_JobQueue_Data))
  Protected *Thread._JobQueue_WorkerInit
  If Worker< = 0
    Worker = 1
  EndIf
  With *JobQueue
    ; Stacks Initialisieren
    \Stack[#Job_LowPriority] = Stack_New()
    \Stack[#Job_NormalPriority] = Stack_New()
    \Stack[#Job_HighPriority] = Stack_New()
    \ReturnBuffer = Stack_New()
    ; Signal-Semaphoren Initialisieren
    \AllJobsFinished = CreateSemaphore()
    \Resume = CreateSemaphore()
    ; Mutexen Initialisieren
    \Mutex = CreateMutex()
    \DelayMutex = CreateMutex()
    ; Sonstiges
    \CleanDelay = 2500 ; Intervall fürs Auto-Aufräumen des Result-Speichers
    \WorkerCount = Worker ; Worker Anzahl
    \Thread = AllocateMemory((\WorkerCount + 1)*SizeOf(_JobQueue_Worker)) ; WorkerInfo - Speicher
    
    For i = 1 To Worker ; Threads starten
      *Thread = AllocateMemory(SizeOf(_JobQueue_WorkerInit))
      *Thread\WorkerNumber = i
      *Thread\JData = *JobQueue
      \Thread\Entry[i]\Number = i
      \Thread\Entry[i]\Mutex = CreateMutex()
      \Thread\Entry[i]\Id = CreateThread(@JQ_Worker(), *Thread) ; Worker starten
    Next
    ; Cleaner starten
    \Cleaner = ET_New(@JQ_JobResultCleaner(), \CleanDelay, *JobQueue)
  EndWith
  ProcedureReturn *JobQueue
EndProcedure

; Gibt eine Job Queue Frei
Procedure JQ_DeInit(*JobQueue._JobQueue_Data, TimeOut = 0)
  Protected Priority, Worker, cTime
  With *JobQueue
    ; Ceaner Stoppen
    If \Cleaner
      ET_Free(\Cleaner)
      \Cleaner = #Null
    EndIf 
    
    \Stop = #True ; Deinit - Flag setzen
    
    _JQ_SendCheckSignal(*JobQueue) ; Worker sollen Flags prüfen
    
    If TimeOut
      cTime = ElapsedMilliseconds()
      While (\Stoped<>\WorkerCount Or \JobDelayed) And ElapsedMilliseconds()-cTime<TimeOut
        Delay(1)
      Wend ; Warten bis alle Threads beendet haben
      If \Stoped<>\WorkerCount
        ProcedureReturn #False
      EndIf
    Else
      While \Stoped<>\WorkerCount Or \JobDelayed
        Delay(1)
      Wend ; Warten bis alle Threads beendet haben
    EndIf
    
    ; DelayMutex freigeben
    FreeMutex(\DelayMutex)
    
    For Worker = 1 To \WorkerCount ; Worker mutexen freigebn
      FreeMutex(*JobQueue\Thread\Entry[Worker]\Mutex)
    Next
    For Priority = #Job_HighPriority To #Job_LowPriority Step -1 ; Stacks freigeben
      While Stack_Count(\Stack[Priority])
        FreeMemory(Stack_Pop(\Stack[Priority]))
      Wend
      Stack_Free(\Stack[Priority])
    Next
    While Stack_Count(\ReturnBuffer) ; ReturnBuffer freigeben
      FreeMemory(Stack_Pop(\ReturnBuffer))
    Wend
    Stack_Free(\ReturnBuffer)
    
    ; Alles andere freigeben
    FreeSemaphore(\AllJobsFinished)
    FreeSemaphore(\Resume)
    FreeMemory(\Thread)
    FreeMutex(\Mutex)
    FreeMemory(*JobQueue)
    ProcedureReturn #True
  EndWith
EndProcedure

; Wartet bis alle Jobs fertig sind
Procedure JQ_WaitUntilFinish(*JobQueue._JobQueue_Data)
  With *JobQueue
    Repeat ; Da sind Delayed Jobs
      LockMutex(\Mutex)
      While TrySemaphore(\AllJobsFinished) : Wend
      \WaitForFinish = #True ; Wartet auf beenden
      UnlockMutex(\Mutex)
      If \JobCount
        WaitSemaphore(\AllJobsFinished)
      EndIf
    Until (Not \JobDelayed) And (Not \JobCount)
  EndWith
EndProcedure

; Wartet bis Job beendet wurde
Procedure JQ_WaitForJob(*JobQueue._JobQueue_Data, *Job._JobQueue_JobInfo)
  Protected Semaphore
  With *JobQueue
    If *Job\State<#Job_Finished ; Ist der Job noch nicht erledigt?
      Semaphore = CreateSemaphore()
      LockMutex(\Mutex) ; Mutex sperren
      *Job\SendFinishSignal = Semaphore ; SignalSemaphore setzen
      UnlockMutex(\Mutex); Mutex freigeben
      WaitSemaphore(Semaphore) ; Auf Finish Signal warten
      FreeSemaphore(Semaphore) ; Semaphore freigeben
    EndIf
  EndWith
EndProcedure

; Entfernt alle fertigen Jobs aus dem Buffer (Rückgabe: Anzahl verbleibende Jobs)
Procedure JQ_ClearReturnBuffer(*JobQueue._JobQueue_Data, RemoveUncheckedJobs = #True)
  Protected *Job._JobQueue_JobInfo
  Protected *NewBuffer
  Protected RtVar = #False
  
  With *JobQueue
    If Not \JobCount ; Nur wenn keine Jobs mehr da sind (zu gefährlich mit verketteten Jobs....)
      *NewBuffer = Stack_New()
      LockMutex(\Mutex) ; Mutex sperren
      While Stack_Count(\ReturnBuffer)
        *Job = Stack_Pop(\ReturnBuffer)
        If (RemoveUncheckedJobs And *Job\State = #Job_Finished) Or (*Job\State =  #Job_Checked) ; Fertige löschen, Austehende wieder speichern
          FreeMemory(*Job)
        Else
          Stack_Push(*NewBuffer, *Job)
        EndIf
      Wend
      Stack_Free(\ReturnBuffer)
      \ReturnBuffer = *NewBuffer
      UnlockMutex(\Mutex) ; Mutex freigeben
      RtVar = #True
    EndIf
  EndWith
  
  ProcedureReturn RtVar
EndProcedure

; Stellt das Clean - Intervall ein (Default: 2500 ms)
Procedure JQ_SetCleanDelay(*JobQueue._JobQueue_Data, Delay = 2500)
  With *JobQueue
    \CleanDelay = Delay
    ET_SetDelay(\Cleaner, \CleanDelay)
  EndWith
EndProcedure

; Pausiert die Jobverarbeitung, wartet auf Wunsch darauf das alles Worker pausiert sind
Procedure JQ_Pause(*JobQueue._JobQueue_Data, WaitUntilConfirmation = #False)
  Protected PausedCount
  With *JobQueue
    If Not \Pause
      \Pause = CreateSemaphore()
      _JQ_SendCheckSignal(*JobQueue)
      
      If WaitUntilConfirmation ; Auf bestätigungen Warten falls Verlang
        Repeat
          WaitSemaphore(\Pause)
          PausedCount+1
        Until PausedCount = \WorkerCount
      EndIf
      
    EndIf
  EndWith
EndProcedure

; Setz die Jobverarbeitung fort
Procedure JQ_Resume(*JobQueue._JobQueue_Data)
  Protected Workers, i
  With *JobQueue
    If \Pause

      FreeSemaphore(\Pause)
      \Pause = #False

      For i = 1 To \WorkerCount
        SignalSemaphore(\Resume)
      Next
    EndIf
  EndWith
EndProcedure

; Fügt einen Job hinzu (Rückgabewert: JobID für JobQueue_JobFinished()/JobQueue_JobReturnValue())
Procedure JQ_Add(*JobQueue._JobQueue_Data, *JobProcedure.JobQueue_Job, Parameter.i, Priority = #Job_NormalPriority, Parent = #Null, MinimumLaunchDelay = #Null, _DelayLaunchInfo = #Null)
  Protected *Job._JobQueue_Job = AllocateMemory(SizeOf(_JobQueue_Job))
  Protected *JobInfo._JobQueue_JobInfo = AllocateMemory(SizeOf(_JobQueue_JobInfo))
  Protected *DelayLaunch._JobQueue_DelayLauncher
  With *JobQueue
    ; Job erstellen
    *Job\Call = *JobProcedure
    *Job\Para = Parameter
    *Job\Info = *JobInfo
    *Job\Parent = Parent
    If Priority = #Job_RealTimePriority
      *Job\RealTime = #True ; RealTime Priorität. ACHTUNG!!!!
    EndIf
    
    If MinimumLaunchDelay ; Verzögerter start
      *DelayLaunch = AllocateMemory(SizeOf(_JobQueue_DelayLauncher))
      *DelayLaunch\Job = *Job
      *DelayLaunch\JobQueue = *JobQueue
      *DelayLaunch\Priority = Priority
      ET_New(@JQ_DelayLauncher(), MinimumLaunchDelay, *DelayLaunch, #True)
      LockMutex(\DelayMutex)
        \JobDelayed+1
      UnlockMutex(\DelayMutex)
      ProcedureReturn *Job\Info
    EndIf
    
    LockMutex(\Mutex) ; Mutex sperren
    
    \JobCount + 1
    If _DelayLaunchInfo ; Infos vom verzögerten Start übernehmen
      LockMutex(\DelayMutex)
        \JobDelayed-1
      UnlockMutex(\DelayMutex)
      FreeMemory(*Job\Info)
      *Job\Info = _DelayLaunchInfo
    EndIf
    
    Stack_Push(\ReturnBuffer, *Job\Info) ; ReturnBuffer speichern
    
    
    ; Job auf den Stack pushen
    If Priority >= #Job_HighestPriority
      Stack_Push(\Stack[#Job_HighPriority], *Job) ; HighPriority, Erste Stelle
    Else
      Stack_PushF(\Stack[Priority], *Job)
    EndIf
    UnlockMutex(\Mutex); Mutex freigeben
    ProcedureReturn *Job\Info
  EndWith
EndProcedure

; gibt die Anzahl der ausstehenden Jobs an
Procedure JQ_Count(*JobQueue._JobQueue_Data)
  Protected Count
  With *JobQueue

    Count = \JobCount

    ProcedureReturn Count
  EndWith
EndProcedure

; gibt die Anzahl der verzögerten Jobs an
Procedure JQ_CountDelayed(*JobQueue._JobQueue_Data)
  Protected Count
  With *JobQueue

    Count = \JobDelayed

    ProcedureReturn Count
  EndWith
EndProcedure

; Rückgabewert des Jobs (Auch aufrufen wenn kein Rückgabewert, da sich der Rückgabestack unnötig auffbläst oder manuel Leeren)
; ACHTUNG!! Nur 1* pro job aufrufen, sobald der Wert abgefragt wurde kümmert sich der Cleaner um die JobInfos
; Nach dem Aufruf dieser Procedure sollte der Job auch mit keiner anderen Procedure mehr verwendet werden.
Procedure JQ_JobReturnValue(*JobQueue._JobQueue_Data, *Job._JobQueue_JobInfo)
  Protected RtVar
  With *JobQueue

    If *Job\State = #Job_Finished
      RtVar = *Job\ReturnValue
      *Job\State = #Job_Checked
    EndIf

    ProcedureReturn RtVar
  EndWith
EndProcedure
; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 133
; Folding = ----
; EnableThread
; EnableXP
; Executable = ..\..\demo.exe
; CPU = 1
; DisableDebugger
; EnableCompileCount = 835
; EnableBuildCount = 1
; EnableExeConstant