; ------------------------------------------------------------------------------------
; Einfaches Logsystem
; Author: Alexander Aigner
; PB 4.4+ x86/x64 Ascii/Unicode
; ------------------------------------------------------------------------------------

XIncludeFile "common\LinkedList.pbi"

; EnableExplicit

Structure Logger_Entry
  Priority.l
  Time.l
  *Message
EndStructure

Structure Logger_Data
  *LogList.LL_Data
  MinViewPriority.l
  *Current.Logger_Entry
  Enable.a
EndStructure

Enumeration
  #Logger_Info
  #Logger_Warning
  #Logger_Error
EndEnumeration

; ------------------------------------------------------------------------------------
;  Erstellt einen neuen Logger
;
;  Rckgabe: Logger - Handle
; ------------------------------------------------------------------------------------
Procedure Logger_New()
  Protected *Logger.Logger_Data = AllocateMemory(SizeOf(Logger_Data))
  
  With *Logger
    \LogList = LL_New()  
    \Enable = #True
  EndWith
  
  ProcedureReturn *Logger  
EndProcedure


; ------------------------------------------------------------------------------------
;  Lscht alle Eintrge im Logger
;  *Logger: Logger Handle
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure Logger_Clear(*Logger.Logger_Data)
  Protected *LogEntry.Logger_Entry
  
  With *Logger
    LL_Reset(\LogList)
    While LL_Next(\LogList)
      *LogEntry = LL_Get(\LogList)
      LL_Delete(\LogList)
      FreeMemory(*LogEntry\Message)
      FreeMemory(*LogEntry)
    Wend
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Gibt den Logger frei
;  *Logger: Logger Handle
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure Logger_Free(*Logger.Logger_Data)
  
  With *Logger
    Logger_Clear(*Logger)
    LL_Free(\LogList)
    FreeMemory(*Logger)
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Aktiviert/Deaktiviert den Logger
;  *Logger: Logger Handle
;  Enable: #False: Logger ist deaktiviert, sonst aktiviert
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure Logger_Enable(*Logger.Logger_Data, Enable = #True)
  
  With *Logger
    \Enable = Enable 
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Erstellt einen neuen Eintrag im Logger
;  *Logger: Logger Handle
;  Eintrag: Log - Nachricht
;  Priority: Prioritt des Eintrags
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure Logger_Log(*Logger.Logger_Data, Message$, Priority = #Logger_Info)
  Protected *LogEntry.Logger_Entry
  
  With *Logger
    If Message$ And \Enable
      LL_Last(\LogList)
      
      *LogEntry = AllocateMemory(SizeOf(Logger_Entry))
      *LogEntry\Time = Date()
      *LogEntry\Priority = Priority
      *LogEntry\Message = AllocateMemory(StringByteLength(Message$) + SizeOf(Character))
      PokeS(*LogEntry\Message, Message$)
      
      LL_Add(\LogList)
      LL_Set(\LogList, *LogEntry)
    EndIf 
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Startet eine Auflistung aller Eintrge im Logger
;  *Logger: Logger Handle
;  MinimumViewPriority: Minimale Prioritt der Eintrge die aufgelistet werden sollen
;
;  Rckgabe: 0
; ------------------------------------------------------------------------------------
Procedure Logger_StartView(*Logger.Logger_Data, MinimumViewPriority = #Logger_Info)
  
  With *Logger
    LL_Reset(\LogList)
    \MinViewPriority = MinimumViewPriority
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Ld den nchsten Eintrag aus dem Logger
;  *Logger: Logger Handle
;
;  Rckgabe: #True wenn Eintrag vorhanden, sonst #False
; ------------------------------------------------------------------------------------
Procedure Logger_ViewNext(*Logger.Logger_Data)
  Protected RtVar, *LogEntry.Logger_Entry
  
  With *Logger
    \Current = #Null
    
    While LL_Next(\LogList)
      *LogEntry = LL_Get(\LogList)
      If *LogEntry\Priority >= \MinViewPriority
        \Current = *LogEntry
        RtVar = #True
        Break
      EndIf
    Wend
  EndWith
  
  ProcedureReturn RtVar
EndProcedure

; ------------------------------------------------------------------------------------
;  Ld die Zeit des aktuellen Eintrags
;  *Logger: Logger Handle
;
;  Rckgabe: Zeit
; ------------------------------------------------------------------------------------
Procedure Logger_GetTime(*Logger.Logger_Data)
  
  With *Logger
    If \Current
      ProcedureReturn \Current\Time  
    EndIf
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Ld die Prioritt des aktuellen Eintrags
;  *Logger: Logger Handle
;
;  Rckgabe: Prioritt
; ------------------------------------------------------------------------------------
Procedure Logger_GetPriority(*Logger.Logger_Data)
  
  With *Logger
    If \Current
      ProcedureReturn \Current\Priority 
    EndIf
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
;  Ld die Nachricht des aktuellen Eintrags
;  *Logger: Logger Handle
;
;  Rckgabe: Nachricht
; ------------------------------------------------------------------------------------
Procedure.s Logger_GetMessage(*Logger.Logger_Data)
  
  With *Logger
    If \Current
      ProcedureReturn PeekS(\Current\Message)
    EndIf
  EndWith
EndProcedure

; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 6
; FirstLine = 4
; Folding = --
; EnableXP
; CPU = 1